#include "disk.h"
#include "file.h"
#include <stdio.h>
#include <string.h>

int stepSc = 8;
int stepTr = 8;
byte *fDir = 0;
int fDirSec = 0;
int fDirTrk = 0;
bool writeDir = true;
bool debug = false;

byte sectorsPerTrack[] = { 0,
  21, 21, 21, 21, 21, 21, 21, 21,
  21, 21, 21, 21, 21, 21, 21, 21,
  21, 19, 19, 19, 19, 19, 19, 19,
  18, 18, 18, 18, 18, 18, 17, 17,
  17, 17, 17
};

long locateOnDisk(byte track, byte sector)
{
  long offs = 0;

  if(track > 17) {
    offs += (256 * 21 * 17);
    if(track > 24) {
      offs += (256 * 19 * 7);
      if(track > 30) {
        offs += (256 * 18 * 6);
        offs += (256 * 17 * (track - 31));
      } else {
        offs += (256 * 18 * (track - 25));
      }
    } else {
      offs += (256 * 19 * (track - 18));
    }
  } else {
    offs += (256 * 21 * (track - 1));
  }
  offs += (256 * sector);

  return offs;
}

void writeC64File(File *c64file,
		  File *d64file,
		  byte* startTrack,
		  byte* startSector,
		  bool raw)
{
  uint sizeInBlocks, blocksFree = 0, i;
  byte *bootBlock = &d64file->data[BootOffset];
  byte *currentBlock, *bam, *dir;
  byte currentTrack, currentSector;
  byte nextTrack, nextSector;
  byte *fData = c64file->data;
  uint fSize = c64file->size;
  uint dirOffset;
  uint blockSize;

  // Helper function..
  void findNextSector() {
    // Find next sector
    currentSector += stepSc;
    currentSector %= sectorsPerTrack[currentTrack];

    if(debug) {
      printf("next t,s = %i, %i -> ", currentTrack, currentSector);
    }

    for(; currentTrack < 36;) {
      bam = bootBlock + (currentTrack * 4);
      for(i = 0; i < sectorsPerTrack[currentTrack]; ++i) {
	if((*(bam + (currentSector / 8) + 1) &
	    (1 << (currentSector & 7))) != 0) {
	  break;
	}
	if(debug) {
	  printf(".");
	}
	currentSector++;
	currentSector %= sectorsPerTrack[currentTrack];
      }
      if((*(bam + (currentSector / 8) + 1) &
	  (1 << (currentSector & 7))) != 0) {
	break;
      }

      currentTrack++;
      currentSector += stepTr;
      currentSector %= sectorsPerTrack[currentTrack];
    }
  }
  // End of helper function..


  if(raw == false) {
    blockSize = 254;
  } else {
    blockSize = 256;
  }

  // First some sanity checks..
  if(fSize == 0) return;

  sizeInBlocks = (fSize + blockSize - 1) / blockSize;
  for(i = 1; i < 36; ++i) {
    if(i != 18) {
      byte b = *(bootBlock + (i * 4));
      blocksFree += b;
    }
  }
  if((sizeInBlocks > blocksFree) && !raw) {
    printf("File \"%s\" too large.\n", c64file->name);
    return;
  }

  // Set starting location on disk
  if(raw) {
    // Verify that the starting sector is free
    if(((*(bootBlock + (*startTrack * 4) + (*startSector / 8) + 1) &
	 (1 << (*startSector & 7))) == 0) ||
       ((*startTrack == 18) && (*startSector == 0))) {
      printf("Writing raw file to already allocated tr, sc: %i, %i.\n",
	     *startTrack, *startSector);
    }
  } else {
    // Verify that the starting sector is free
    if(((*(bootBlock + (*startTrack * 4) + (*startSector / 8) + 1) &
	 (1 << (*startSector & 7))) == 0) ||
       ((*startTrack == 18) && (*startSector == 0))) {

      // Find first free block
      for(i = 1; i < 36; ++i) {
	if((*(bootBlock + (i * 4)) != 0) && 
	   i != 18) {
	  bam = bootBlock + (i * 4);
	  *startTrack = i;
	  
	  for(i = 0; i < sectorsPerTrack[*startTrack]; ++i) {
	    if((*(bam + (i / 8) + 1) &
		(1 << (i & 7))) != 0) {
	      *startSector = i;
	      break;
	    }
	  }
	  break;
	}
      }
    }
  }

  currentTrack = *startTrack;
  currentSector = *startSector;

  // Copy data to block(s)
  while(fSize > blockSize) {
    bam = bootBlock + (currentTrack * 4);

    // Locate block in d64file
    currentBlock = d64file->data + locateOnDisk(currentTrack, currentSector);

    // Write one block
    for(i = 0; i < blockSize; ++i, --fSize) {
      if(raw)
	*(currentBlock + i) = *fData++;
      else
	*(currentBlock + i + 2) = *fData++;
    }
    *(bam + (currentSector / 8) + 1) &=
      (255 - (1 << (currentSector & 7)));
    *bam -= 1;

    if(debug) {
      printf("t,s = %i, %i. ", currentTrack, currentSector);
      printf("bam = %x %x %x %x. ", *bam, *(bam+1), *(bam+2), *(bam+3));
    }

    findNextSector();

    if(currentTrack > 35) {
      printf("Write Error.\n");
      return;
    }

    if(debug) {
      printf("%i, %i.\n", currentTrack, currentSector);
    }

    if(!raw) {
      *(currentBlock) = currentTrack;
      *(currentBlock + 1) = currentSector;
    }
  }

  // Write last sector
  currentBlock = d64file->data + locateOnDisk(currentTrack, currentSector);
  for(i = 0; fSize > 0; ++i, --fSize) {
    if(raw)
      *(currentBlock + i) = *fData++;
    else
      *(currentBlock + i + 2) = *fData++;
  }
  if(!raw) {
    *(currentBlock) = 0;
    *(currentBlock + 1) = i + 1;
  }
  bam = bootBlock + (currentTrack * 4);
  *(bam + (currentSector / 8) + 1) &=
    (255 - (1 << (currentSector & 7)));
  *bam -= 1;

  if(debug) {
    printf("t,s = %i, %i. ", currentTrack, currentSector);
    printf("bam = %x %x %x %x.\n", *bam, *(bam+1), *(bam+2), *(bam+3));
  }

  printf("Wrote \"%s\" to tr, sc: %i, %i.\n",
	 c64file->name, *startTrack, *startSector);


  // Find next preferred tr/sc for next file..
  findNextSector();
  nextTrack = currentTrack;
  nextSector = currentSector;


  // Add dir entry
  if(!raw) {

    // Add fast dir entry
    if(fDir != 0) {
      *(fDir++) = *startTrack;
      *(fDir++) = *startSector;

      // Allocate the fast dir in BAM
      bam = bootBlock + (fDirTrk * 4);
      *(bam + (fDirSec / 8) + 1) &=
	(255 - (1 << (fDirSec & 7)));
      *bam -= 1;
    }
  }

  if(!raw && writeDir) {

    currentTrack = *(bootBlock);
    currentSector = *(bootBlock + 1);
    bam = bootBlock + (18 * 4);

    for(;;) {
      currentBlock = d64file->data + locateOnDisk(currentTrack, currentSector);
      for(dirOffset = 2; dirOffset < 256; dirOffset += 32) {
	if(*(currentBlock + dirOffset) == 0) {
	  break;
	}
      }
      if(dirOffset < 256) {
	dir = currentBlock + dirOffset;
	break;
      } else {
	if(*(currentBlock) == 18) { // Next sector
	  currentSector = *(currentBlock + 1);
	}
      }

    }

    *(dir) = 0x82;
    *(dir + 1) = *startTrack;
    *(dir + 2) = *startSector;
    for(i = 0; i < 16; ++i) {
      if(strlen(c64file->name) > i) {
	byte c = c64file->name[i];
	if(c >= 0x40 && c < 0x80)
	  c &= 0x5f;
	*(dir + 3 + i) = c;
      } else {
	*(dir + 3 + i) = 0xa0;
      }
    }
    *(dir + 28) = sizeInBlocks % 256;
    *(dir + 29) = sizeInBlocks / 256;
  }

  *startTrack = nextTrack;
  *startSector = nextSector;
}

void writeC64FileFirstFree(File *c64file,
			   File *d64file)
{
  static byte prefTrack = 1;
  static byte prefSector = 0;

  writeC64File(c64file, d64file, &prefTrack, &prefSector, false);
}

int main(int argc, char * argv[])
{
  File d64File;
  File c64File;
  inputstate state;
  int i;
  byte prefTrack, prefSect;
  bool isRaw = false;

  if(argc <= 1) {
    printf("Usage:\n"
	   "disk [-name <name>]: Name of disk image (default=\"out.d64\").\n"
	   "     [-h <text>]: Set disk header, max 21 chars.\n"
           "     [-id <num>]: Put id (0-9) last byte on boot block.\n"
	   "     [-w <name>*]: Write file(s).\n"
	   "     [-wp <loc> <name>]: Write file to preferred track and sector\n"
	   "     [-wx <loc> <name>]: Write raw file to track and sector.\n"
	   "     [-ss <num>]: Set sector step (interleave) (-20 to 20) (default=8).\n"
	   "     [-st <num>]: Set extra sector step when changing track (default=8).\n"
	   "     [-fdir <loc>]: Put fast dir at track and sector.\n"
	   "     [-nodir]: Don't write dir entry.\n"
	   "     [-dir]: Do write dir entry.\n"
	   "     [-debug]: Output some debug info..\n");
    return;
  }


  // First check for .d64 name and debug flag.
  state = idle;
  d64File.name = "out.d64";
  for(i = 1; i < argc; ++i) {
    switch(state) {
    case idle:
      if(strcmp(argv[i], "-name") == 0)
	state = diskName;
      if(strcmp(argv[i], "-debug") == 0)
	debug = true;
      break;
    case diskName:
      d64File.name = argv[i];
      state = idle;
      break;
    }
  }

  // Exit if syntax error
  if(state != idle) {
    printf("Error(1): Syntax error.\n");
    freeFile(&d64File);
    return -1;
  }


  // Create empty .d64
  createD64(&d64File);


  // This loop counts dir-entries, so that we know
  // how many blocks will be used by the dir.
  // Any unused block on track 18 may be used for philez.
  state = idle;
  writeDir = true;
  int numberOfDirEntries = 0;
  for(i = 1; i < argc; ++i) {

    if(state == writeFileN) {
      if(strncmp(argv[i], "-", 1) == 0) {
	state = idle;
      } else {
	state = writeFile1;
      }
    }

    switch(state) {
    case idle:
      if(strcmp(argv[i], "-nodir") == 0) {
	writeDir = false;
      }
      if(strcmp(argv[i], "-dir") == 0) {
	writeDir = true;
      }
      if(strcmp(argv[i], "-name") == 0)	state = diskName;
      if(strcmp(argv[i], "-ss") == 0) state = stepSector;
      if(strcmp(argv[i], "-st") == 0) state = stepTrack;
      if(strcmp(argv[i], "-id") == 0) state = putId;
      if(strcmp(argv[i], "-h") == 0) state = header;
      if(strcmp(argv[i], "-fdir") == 0) state = fastDir;
      if(strcmp(argv[i], "-w") == 0) state = writeFile1;
      if(strcmp(argv[i], "-wp") == 0) {
	isRaw = false;
	state = writeFileTS1;
      }
      if(strcmp(argv[i], "-wx") == 0) {
	isRaw = true;
	state = writeFileTS1;
      }
      break;

    case writeFile1:
      if(writeDir) {
	numberOfDirEntries++;
      }
      state = writeFileN;
      break;

    case writeFileTS1:
      state = writeFileTS2;
      break;

    case writeFileTS2:
      if(!isRaw && writeDir) {
	numberOfDirEntries++;
      }
      state = idle;
      break;

    case diskName:
    case fastDir:
    case stepSector:
    case stepTrack:
    case putId:
    case header:
      state = idle;
      break;
    }
  }

  // Exit if syntax error
  if(state != idle && state != writeFileN) {
    printf("Error(2): Syntax error.\n");
    freeFile(&d64File);
    freeFile(&c64File);
    return -1;
  }


  // Calculate number of sectors needed for dir
  int numberOfDirSectors = (numberOfDirEntries + 7) / 8;
  if(debug) {
    printf("%i dir entries using %i sectors.\n",
	   numberOfDirEntries, numberOfDirSectors);
  }

  // Pre-allocate sectors for directory
  byte *bam = d64File.data + locateOnDisk(18, 0) + (18 * 4);
  byte currentSector = 0;
  byte nextSector = 1;
  while(numberOfDirSectors > 0) {

    byte *currentBlock = d64File.data + locateOnDisk(18, currentSector);

    // Link to next sector
    *(currentBlock) = 18;
    *(currentBlock + 1) = nextSector;

    // Allocate sector in BAM
    *(bam + (nextSector / 8) + 1) &=
      (255 - (1 << (nextSector & 7)));
    *bam -= 1;

    // Step sector
    currentSector = nextSector;
    nextSector = (nextSector + 3) % 19;

    numberOfDirSectors--;
  };



  // Here it is..
  // This loop writes files and dir entries to the .d64.
  // ..and takes care of sector stepping and everything..
  state = idle;
  writeDir = true;
  for(i = 1; i < argc; ++i) {
    switch(state) {
    case idle:
    case writeFileN:
      if(strcmp(argv[i], "-name") == 0) {
	state = diskName;
	break;
      }
      if(strcmp(argv[i], "-debug") == 0) {
	// Do nothing
	break;
      }
      if(strcmp(argv[i], "-w") == 0) {
	state = writeFile1;
	break;
      }
      if(strcmp(argv[i], "-wp") == 0) {
	isRaw = false;
	state = writeFileTS1;
	break;
      }
      if(strcmp(argv[i], "-wx") == 0) {
	isRaw = true;
	state = writeFileTS1;
	break;
      }
      if(strcmp(argv[i], "-ss") == 0) {
	state = stepSector;
	break;
      }
      if(strcmp(argv[i], "-st") == 0) {
	state = stepTrack;
	break;
      }
      if(strcmp(argv[i], "-fdir") == 0) {
	state = fastDir;
	break;
      }
      if(strcmp(argv[i], "-nodir") == 0) {
	writeDir = false;
	state = idle;
	break;
      }
      if(strcmp(argv[i], "-dir") == 0) {
	writeDir = true;
	state = idle;
	break;
      }
      if(strcmp(argv[i], "-id") == 0) {
	state = putId;
	break;
      }
      if(strcmp(argv[i], "-h") == 0) {
	state = header;
	break;
      }

      if(state == writeFileN) {
	state = writeFile1;
	// and fall through to next case-clause!!
	// Aha.. it works only if case is right after.
	// No new compare is made, case is just an entry point..
      } else {
	break;
      }

    case writeFile1:
      c64File.name = argv[i];
      if(!readFile(&c64File)) {
	printf("Error: Open file \"%s\", aborting.\n", c64File.name);
	return -1;
      }
      writeC64FileFirstFree(&c64File, &d64File);
      state = writeFileN;
      break;

    case writeFileTS1:
    case fastDir: {
      uint temp = atoi(argv[i]);
      prefTrack = temp / 100;
      prefSect = temp % 100;
      if(prefTrack > 40 ||
	 prefSect > 21) {
	printf("Error: Illegal Track or Sector: %i, %i. Aborting.\n",
	       prefTrack, prefSect);
	return -1;
      }
      if(state == fastDir) {
	fDirTrk = prefTrack;
	fDirSec = prefSect;
	fDir = d64File.data + locateOnDisk(prefTrack, prefSect);
	state = idle;
      } else {
	state = writeFileTS2;
      }
      break;
    }

    case writeFileTS2:
      c64File.name = argv[i];
      if(!readFile(&c64File)) {
	printf("Error: Open file \"%s\", aborting.\n", c64File.name);
	return -1;
      }
      writeC64File(&c64File, &d64File, &prefTrack, &prefSect, isRaw);
      state = idle;
      break;

    case stepSector: {
      int temp = atoi(argv[i]);
      if((temp > -21) && (temp < 21))
	stepSc = temp;
      else
	printf("Illegal step: %i.\n", temp);
      state = idle;
      break;
    }

    case stepTrack: {
      int temp = atoi(argv[i]);
      if((temp > -21) && (temp < 21))
	stepTr = temp;
      else
	printf("Illegal step: %i.\n", temp);
      state = idle;
      break;
    }

    case putId: {
      d64File.data[BootOffset + 0xff] = *(argv[i]);
      state = idle;
      break;
    }

    case header: {
      int ind = 0x90;
      int len = strlen(argv[i]);
      if(len > 21)
	len = 21;
      int c;
      for(c = 0; c < len; c++) {
	d64File.data[BootOffset + ind] = *((argv[i] + c));
	if(++ind == 0xa0)
	  ind = 0xa2;
      }
      state = idle;
      break;
    }

    case diskName:
      state = idle;
      break;
    }
  }

  // Exit if syntax error
  if(state != idle && state != writeFileN) {
    printf("Error(3): Syntax error.\n");
    freeFile(&d64File);
    freeFile(&c64File);
    return -1;
  }

  if(!writeFile(&d64File)) {
    printf("Error: Write file \"%s\", aborting.\n", d64File.name);
  }

  freeFile(&d64File);
  freeFile(&c64File);

  return 0;
}
