#include "file.h"
#include <stdlib.h>

void createD64(File *aFile)
{
  int i;
  byte bootBlock[] = {
    0x12, 0x01, 0x41, 0x00, // TrackSector link + id
    0x15, 0xff, 0xff, 0x1f, // Track 1 BAM
    0x15, 0xff, 0xff, 0x1f, // Track 2
    0x15, 0xff, 0xff, 0x1f, // ..
    0x15, 0xff, 0xff, 0x1f,
    0x15, 0xff, 0xff, 0x1f,
    0x15, 0xff, 0xff, 0x1f,
    0x15, 0xff, 0xff, 0x1f,
    0x15, 0xff, 0xff, 0x1f,
    0x15, 0xff, 0xff, 0x1f,
    0x15, 0xff, 0xff, 0x1f,
    0x15, 0xff, 0xff, 0x1f,
    0x15, 0xff, 0xff, 0x1f,
    0x15, 0xff, 0xff, 0x1f,
    0x15, 0xff, 0xff, 0x1f,
    0x15, 0xff, 0xff, 0x1f,
    0x15, 0xff, 0xff, 0x1f,
    0x15, 0xff, 0xff, 0x1f,
    0x11, 0xfc, 0xff, 0x07, // Track 18 (BootTrack)
    0x13, 0xff, 0xff, 0x07,
    0x13, 0xff, 0xff, 0x07,
    0x13, 0xff, 0xff, 0x07,
    0x13, 0xff, 0xff, 0x07,
    0x13, 0xff, 0xff, 0x07,
    0x13, 0xff, 0xff, 0x07,
    0x12, 0xff, 0xff, 0x03,
    0x12, 0xff, 0xff, 0x03,
    0x12, 0xff, 0xff, 0x03,
    0x12, 0xff, 0xff, 0x03,
    0x12, 0xff, 0xff, 0x03,
    0x12, 0xff, 0xff, 0x03,
    0x11, 0xff, 0xff, 0x01,
    0x11, 0xff, 0xff, 0x01,
    0x11, 0xff, 0xff, 0x01,
    0x11, 0xff, 0xff, 0x01,
    0x11, 0xff, 0xff, 0x01, // ..and Track 35
    '-', ' ', 'B', 'O', 'O', 'Z', 'E', ' ',
    'D', 'E', 'S', 'I', 'G', 'N', ' ', '-',
    0xa0, 0xa0, '1', '3', '3', '7', '!',
    0xa0, 0xa0, 0xa0, 0xa0
  };

  aFile->size = d64Size;
  aFile->data = (byte *)malloc(d64Size);

  // Init d64.
  for(i = 0; i < d64Size; ++i) {
    aFile->data[i] = 0;
  }
  for(i = 1; i < d64Size; i += 256) {
    aFile->data[i] = 0xff;
  }
  for(i = 0; i < 0xab; ++i) {
    aFile->data[BootOffset + i] = bootBlock[i];
  }
}

void freeFile(File *aFile)
{
  free(aFile->data);
}

bool readFile(File *aFile)
{
  FILE *fp = NULL;
  struct stat fileStatus;

  if(stat(aFile->name, &fileStatus) == -1) {
    return false;
  }
  aFile->size = fileStatus.st_size;

  fp = fopen(aFile->name, "rb");
  if(fp == NULL) {
    return false;
  }

  aFile->data = (byte *)malloc(aFile->size);
  if(aFile->data == NULL) {
    fclose(fp);
    return false;
  }

  if(fread(aFile->data, 1, aFile->size, fp) != aFile->size) {
    free(aFile->data);
    fclose(fp);
    return false;
  }

  fclose(fp);
  return true;
}

bool writeFile(File *aFile)
{
  FILE *fp = NULL;

  fp = fopen(aFile->name, "wb");
  if(fp == NULL) {
    return false;
  }

  if(fwrite(aFile->data, 1, aFile->size, fp) != aFile->size) {
    fclose(fp);
    return false;
  }

  fclose(fp);
  return true;
}
